/*****************************************************************************
* This program accepts 2 files as command line arguments.  It creates 3 output
*   files.  One named unique1, one named unique2, and a 3rd 
*   named common.  Unique1 contains all of the lines from the first
*   file that aren't found in the second.  Unique2 contains all of the
*   lines from the second file that aren't found in the first, and the common
*   contains all of the lines that are common to the 2 files.
*
* The 2 input files must be test files.  They do not have to have the same
*   number of lines.  Neither do they have to have the lines in the same 
*   order.
*
* I created this program for comparing system files, such as autoexec.bat,
*   config.sys, win.ini, etc.  Often, a program will work on one system, but
*   not another, and to all appearances, the systems are identical.  Hopefully,
*   this will aid in the troubleshooting.
*
* 06 Dec. 1996 - Started work.  Created fileline structure to keep track of
*                  each line in the input files.  The lines will be kept in a
*                  linked list for each file.  Created code to open all of the
*                  files, and allocate all of the buffers.  Also, to close all
*                  files, and free all buffers.  Created code to read file1 and
*                  file2 into in memory singly linked list.  Created the code
*                  to do the searching.  Program now works as advertised.
*
* 07 Dec. 1996 - Added the variables bytes1 and bytes2 to track memory usage
*                  for the file processing.  Also added the variables nlines1 
*                  and nlines2 to hold the number of lines in files 1 & 2
*                  respectively.  All of these variables are strictly for
*                  informational purposes only.  They are used at the end
*                  for the status messages.  It will be interesting to keep
*                  an eye on the memory usage.
*              - Cleaned up the error handling and returned errorlevels.  Added
*                  an errorlevel variable, and then set it for each error, and 
*                  put in a goto to the EXIT label.  After the exit label, the
*                  file handles are closed if they aren't NULL, and all allocated
*                  pointers are freed, once again if they aren't NULL.  Finally,
*                  the status messages are printed, and the errorlevel is returned.
*              - Created the function free_fileline_list to free all of the memory
*                  associated with a linked list of fileline structures.  It 
*                  takes a pointer to a filelist structure, frees all of its
*                  memory, and sets it to NULL.
*              - Created the function alloc_fileline_list to create the linked
*                  list of fileline structures from a text file.  This function
*                  accepts the filename, a pointer to an integer to store the number
*                  of bytes used by the list, and a pointer to an integer to store
*                  the number of lines in the file.  It returns a NULL if there is
*                  any error.  Just as in main I used an EXIT label, in this function
*                  I used an ALLOC_FILELINE_LIST_EXIT label to facilitate error
*                  processing.
*              - Created  the function is_line_notblank.  This function accepts a 
*                  NULL terminated string, and returns 1 if the string contains at
*                  least 1 non whitespace or non control character, 0 otherwise.  Used
*                  this function to stop processing of blank lines in both files.
*              - Changed the strcmp() call that compares the two current lines to a
*                  stricmp() call.  This says lines are matches if they say the same
*                  thing but in different styles/mixes of capitalization.  Also added
*                  header lines to unique1 and unique2, and several header lines to
*                  common.
*              - Changed the name from filecomp to fintrsct, short for file intersection,
*                  like the intersection operation from set theory.
*              - Added Bionic Fish Software Foundry copyright message.
*
* 09 Dec. 1996 - For some reason, under NT4.0 the code in the EXIT section of main
*                  that tested to see if lin1!=NULL and then free(lin1) doesn't work.
*                  It appears to a code generation or timing thing.  So I changed it 
*                  from:  if(lin1!=NULL) free(lin1); to:
*                  if(lin1!=NULL) free_fileline_list(lin1); and this works.  Also 
*                  changed the code that does the same thing for lin2.
*
* 07 Sep. 1998 - The code I thought was fixed has now blown up on an NT 4.0 SP3 
*                  system.  Changing the cleanup code to use the errorlevel 
*                  variable as an indication of what to do.  This is my "standard"
*                  cleanup system (first used in usage.c).
*              - Changed date and release from 07 Dec. 1996 and V1R0 to 07 Sep. 1998
*                  and V1R1.
*
* Errorlevels: 255 - Invalid or not enough command line arguments
*              252 - Error opening unique1 for write
*              251 - Error opening unique2 for write
*              250 - Error opening common for write
*              248 - Error allocating lin1
*              247 - Error allocating lin2
*
* 32 Bit version
* VC 4.0 - Compile:  cl /nologo /W3 /D "WIN32" /D "CONSOLE" /c fintrsct.c
* VC 4.0 - Link:     link /NOLOGO /MACHINE:IX86 /SUBSYSTEM:CONSOLE 
*                    /OUT:fintrsct.exe fintrsct.obj
*
* 16 bit version
* QC 2.5 - Compile: cl /AL /J /Zp /Oailt /W3 /DLINT_ARGS /Gs /c fintrsct.c
* QC 2.5 - Link:    qlink /E /NOI /PACKCODE /SE:256 fintrsct,fintrsct,,,,
*****************************************************************************/
#include <stdio.h>                               /* Standard C */
#include <stdlib.h>                              /* Standard C */
#include <string.h>                              /* Standard C */
#include <ctype.h>                               /* Standard C */

#define MAXLEN 2048                              /* Longest allowable input line */

struct fileline {                                /* Keep track of one line */
  unsigned long    linenum;                      /* Line number, 0 based */
  unsigned int     common;                       /* 0=unique line, 1=common line */
  unsigned char   *line;                         /* Actual text of the line */
  struct fileline *nextline;                     /* Pointer to next line, NULL = end */
};

struct fileline *alloc_fileline_list(unsigned char *file, \
                                     unsigned long *bytes, unsigned long *linctr) {
  char            *inbuf;                        /* Input buffer */
  FILE            *inf;                          /* Input file */
  struct fileline *ret;                          /* First line in file */
  struct fileline *next;                         /* Next fileline structure */

  (*bytes)=0;                                    /* Set number of bytes used to 0 */
  (*linctr)=0;                                   /* Set line counter to 0 */

  inf=fopen(file,"r");                           /* Open file for input */
  if(inf==NULL) {                                /* If error */
    ret=NULL;                                    /* Return a NULL */    
    goto ALLOC_FILELINE_LIST_EXIT;               /* Exit the function */
  }

  inbuf=(unsigned char *)calloc(MAXLEN,sizeof(unsigned char));
  if(inbuf==NULL) {                              /* If error */
    ret=NULL;                                    /* Return a NULL */
    goto ALLOC_FILELINE_LIST_EXIT;               /* Exit the function */
  }

  ret=(struct fileline *)calloc(1,sizeof(struct fileline));
  (*bytes)+=sizeof(struct fileline);             /* Starting size of 1 structure */

  if(ret!=NULL) {                                /* If ret got allocated */
    next=ret;                                    /* At beginning, next = ret */
    while(fgets(inbuf,MAXLEN,inf)!=NULL) {       /* Loop as long as EOF not reached */
      next->linenum=*linctr;                     /* Set line number */
      next->common=0;                            /* Set common flag to 0 */
      next->line=(unsigned char *)calloc(strlen(inbuf)+1,sizeof(unsigned char));
      (*bytes)+=strlen(inbuf)+1;                 /* Add the memory used for the line string */
      strcpy(next->line,inbuf);                  /* Copy the data */
      next->nextline=(struct fileline *)calloc(1,sizeof(struct fileline));
      (*bytes)+=sizeof(struct fileline);         /* Add the memory used for the next line */
      next=next->nextline;                       /* Goto next line struct */    
      (*linctr)++;                               /* Increment ctr */
      memset(inbuf,MAXLEN,'\0');                 /* Zero inbuf */
    }
    next->nextline=NULL;                         /* Terminate the chain */
  }

  ALLOC_FILELINE_LIST_EXIT:                      /* Function exit label */
  if(inf!=NULL)                                  /* If there was no error */
    fclose(inf);                                 /* Close the input file */
  if(inbuf!=NULL)                                /* If there was no error */
    free(inbuf);                                 /* Free the input buffer */
  return(ret);                                   /* Return the fileline linked list */
}

void free_fileline_list(struct fileline *lines) {
  struct fileline *cur;                          /* Current line */

  while(lines->nextline!=NULL) {                 /* Free all parts of the chain */
    cur=lines;                                   /* Set current line */
    lines=lines->nextline;                       /* Point lines to next line */
    free(cur->line);                             /* Free the line string */
    free(cur);                                   /* Free the line record */
  }
  free(lines);                                   /* Free the last record */
  lines=NULL;                                    /* Set lines to NULL */
}

unsigned int is_line_notblank(unsigned char *line) {
  unsigned char *cur;                            /* Current character in string */
  unsigned int   fnd=0;                          /* Non space or control found */

  for(cur=line; *cur!='\0'; cur++) {             /* Step thru the string */
    if(fnd==0)                                   /* As long as fnd is 0 */
      if((isspace(*cur))||(iscntrl(*cur)))       /* If a whitespace or a control char */
        fnd=0;                                   /* Fnd remains at 0 */
      else                                       /* Otherwise */
        fnd=1;                                   /* Fnd gets set to 1 */
  }

  return(fnd);                                   /* fnd=1, not blank - fnd=0,blank */
}

void main(int argc, char *argv[]) {
  unsigned int     fnd;                          /* Found flag, 0=not found, 1=found */
  unsigned int     errorlevel=0;                 /* Errorlevel returned by this program */
  unsigned long    nlines1=0;                    /* Number of lines in file1 */
  unsigned long    nlines2=0;                    /* Number of lines in file2 */
  unsigned long    bytes1=0;                     /* Bytes used to handle file1 */
  unsigned long    bytes2=0;                     /* Bytes used to handle fiel2 */
  FILE            *un1=NULL;                     /* 1st output file, unique1 */
  FILE            *un2=NULL;                     /* 2nd output file, unique2 */
  FILE            *com=NULL;                     /* 3rd output file, common */
  struct fileline *lin1=NULL;                    /* Pointer to first line of file 1 */
  struct fileline *lin2=NULL;                    /* Pointer to first line of file 2 */
  struct fileline *next1;                        /* Pointer to next line of file 1 */
  struct fileline *next2;                        /* Pointer to next line of file 2 */

  if(argc<3) {                                   /* If file1 and file2 not spec'd */
    printf("USAGE :: fintrsct file1 file2\n");   /* Print usage message */
    printf("\n");
    printf(" file1 : Text file - input - filename\n");
    printf(" file2 : Text file - input - filename\n");
    printf("\n");
    printf(" Creates : unique1 - Text file - output - lines unique to file1\n");
    printf("           unique2 - Text file - output - lines unique to file2\n");
    printf("           common  - Text file - output - lines common to file1 and file2\n");
    printf("\n");
    printf(" fintrsct - File INTeRSeCTion.  Intersection as in the intersection of 2 sets\n");
    errorlevel=255;                              /* Errorlevel 255 - invalid args */
    goto EXIT;                                   /* Exit with errorlevel */
  }

  un1=fopen("unique1","w");                      /* Open unique1 for write */
  if(un1==NULL) {                                /* If error */
    errorlevel=252;                              /* Errorlevel 252 - Error opening un1 */
    goto EXIT;                                   /* Exit with errorlevel */
  }

  un2=fopen("unique2","w");                      /* Open unique2 for write */
  if(un2==NULL) {                                /* If error */
    errorlevel=251;                              /* Errorlevel 251 - Error opening un2 */
    goto EXIT;                                   /* Exit with errorlevel */
  }

  com=fopen("common","w");                       /* Open common for write */
  if(com==NULL) {                                /* If error */
    errorlevel=250;                              /* Errorlevel 250 - Error opening un2 */
    goto EXIT;                                   /* Exit with errorlevel */
  }

  lin1=alloc_fileline_list(argv[1],&bytes1,&nlines1);
  if(lin1==NULL) {                               /* If error */
    errorlevel=248;                              /* Errorlevel 248 - Error creating lin1 */
    goto EXIT;                                   /* Exit with errorlevel */
  }

  lin2=alloc_fileline_list(argv[2],&bytes2,&nlines2);
  if(lin2==NULL) {                               /* If error */
    errorlevel=247;                              /* Errorlevel 247 - Error creating lin2 */
    goto EXIT;                                   /* Exit with errorlevel */
  }

  fprintf(un1,"Lines unique to %s\n\n",argv[1]); /* Unique1 header */
  fprintf(un2,"Lines unique to %s\n\n",argv[2]); /* Unique2 header */
  /* Common header                                                 */  
  fprintf(com,"Lines common to both %s and %s\n",argv[1],argv[2]);
  fprintf(com,"1st line in set is from %s\n",argv[1]);
  fprintf(com,"2nd lines on in set are matching lines from %s\n",argv[2]);
  fprintf(com,"Sets are delineated by blank lines\n");

  next1=lin1;                                    /* At beginning, next1 = lin1 */
  while(next1->nextline!=NULL) {                 /* Step thru each line in file1 */
    if(is_line_notblank(next1->line)==1) {       /* If the line in file1 isn't blank */
      fnd=0;                                     /* Set found flag to not found */
      next2=lin2;                                /* Set next2 = lin2 */
      while(next2->nextline!=NULL) {             /* Loop thru each line in file2 */
        if(stricmp(next1->line,next2->line)==0) {/* If the lines are identical, ignoring case */
          if(fnd==0)                             /* And its the first match in file2 */
            fprintf(com,"\n%ld\t%s",next1->linenum,next1->line);
          fprintf(com,"%ld\t%s",next2->linenum,next2->line);
          next2->common=1;                       /* Set common flag on file2 record */
          fnd=1;                                 /* Set found flag to found */
        }
        next2=next2->nextline;                   /* Goto next line in file2 */
      }
      if(fnd==0)                                 /* If file1 line not found in file2 */
        fprintf(un1,"%ld\t%s",next1->linenum,next1->line);
    }
    next1=next1->nextline;                       /* Goto next line in file1 */
  }

  next2=lin2;                                    /* At beginning, next2 = lin2 */
  while(next2->nextline!=NULL) {                 /* Step thru each line in file2 */
    if((next2->common==0)&&(is_line_notblank(next2->line)==1)) /* If line is not common and not blank */
      fprintf(un2,"%ld\t%s",next2->linenum,next2->line);
    next2=next2->nextline;                       /* Goto next line in file2 */
  }

  EXIT:                                          /* Exit label - free and return errorlevel */
  switch(errorlevel) {
    case   0: printf("%ld bytes used to handle %ld lines in %s\n",bytes1,nlines1,argv[1]);
              printf("%ld bytes used to handle %ld lines in %s\n",bytes2,nlines2,argv[2]);
              free_fileline_list(lin2);          /* Free lin2 if required */
    case 247: free_fileline_list(lin1);          /* Free lin1 if required */
    case 248: fclose(com);                       /* Close common */
    case 250: fclose(un2);                       /* Close unique2 */
    case 251: fclose(un1);                       /* Close unique1 */
    case 252:
    case 255: printf("\nFINTRSCT for WinNT, Win95 & WFW DOS Windows, and MS-DOS\n");
              printf("By: Paul Trout, Bionic Fish Software Foundry\n");
              printf("07 Sep. 1998 - V1R1\n");
              exit(errorlevel);
  }
}
